"""
FileName: utils.py
Description: 
Time: 2020/9/11 17:42
Project: CiLog
Author: Shurui Gui
"""
import re
import sys
import copy


def construct_mark(left_string: str):
    mark = f'[{left_string[0].lower()}{left_string[0].upper()}]'
    left_string = left_string[1:]

    if left_string:
        mark += f'({construct_mark(left_string)})?'

    return mark


def str2mark(string):

    return f'#{construct_mark(string)}#'


class CiLogStdOut(object):

    def __init__(self, logger):
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        sys.stdout = self
        sys.stderr = self
        self.logger = logger
        self.buffer = ''

    def write(self, string):

        self.buffer += string

    def flush(self):
        if not self.buffer:
            return

        buffer = self.buffer
        self.buffer = ''

        for level in ['debug', 'info', 'important', 'table_fromlist', 'warning', 'error', 'critical', 'mail']:
            match = re.match(str2mark(level), buffer)
            if match:
                getattr(self.logger, level)(buffer[match.span()[1]:])
                return

        self.logger.origin(buffer)
        return





"""
FileName: scripts.py
Description: 
Time: 2020/9/12 12:06
Project: CiLog
Author: Shurui Gui
"""

import argparse
import subprocess
import signal
import sys, os
from cilog.api import create_logger, json_mail_setting
import json
import shutil
from definitions import ROOT_DIR



def signal_process():
    os.setsid()
    # signal.signal(signal.SIGHUP, signal.SIG_IGN)

parser = argparse.ArgumentParser()
parser.add_argument('-c', type=str, default='~/.cilog/config.json', help='Config file')
parser.add_argument('-r', type=str, default='./cilog.log', help='Output redirection')
parser.add_argument('-e', action='store_true', default=False, help='Set enable_mail to True.')
parser.add_argument('cmd', nargs='+')
args = parser.parse_known_args()

config_filename = os.path.abspath(os.path.expanduser(args[0].c))
default_config = os.path.abspath(os.path.expanduser('~/.cilog/config.json'))
redirect_filename = os.path.abspath(os.path.expanduser(args[0].r))

print(f'Output -> {redirect_filename}')

if not os.path.exists(config_filename):
    if config_filename == default_config:
        os.makedirs(os.path.dirname(default_config))

        shutil.copy(os.path.join(ROOT_DIR, 'config.json'),
                    default_config)
        print(f'\nDefault config file does not exist.\n'
              f'-------------------------------------\n'
              f'{default_config} is automatically added, please configure it!')
        exit(0)
    else:
        raise FileNotFoundError(f'Config file {config_filename} Not Found.')


with open(config_filename, 'r') as f:
    config = json.load(f)
    logger = create_logger(name=config['log_name'], file=args[0].r, use_color=False, enable_mail=args[0].e,
                           mail_setting=json_mail_setting(config['mail_setting']))


cmd = args[0].cmd + args[1]

process = subprocess.Popen(cmd, preexec_fn=signal_process, stdout=open(redirect_filename, 'a'), stderr=subprocess.STDOUT)

process.wait()
logger.mail(f'Task {" ".join(cmd)} completed.')









"""
FileName: handler.py
Description: 
Time: 2020/9/11 16:40
Project: CiLog
Author: Shurui Gui
"""
import logging, os, shutil, time


class CustomFileHandler(logging.FileHandler):

    def __init__(self, filename, mode='a', encoding=None, delay=False):
        file = os.path.abspath(os.fspath(filename))
        file_path, file_name_ext = os.path.split(file)
        if os.path.exists(file):
            file_name, file_ext = os.path.splitext(file_name_ext)
            f = open(file, 'r', encoding=encoding)
            if len(f.readlines()) > 1e4:
                f.close()
                shutil.copy(file,
                            os.path.join(file_path,
                                         f'{file_name}-{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}{file_ext}'))
                mode = 'w'
        if not os.path.exists(file_path):
            os.makedirs(file_path)
        super().__init__(filename, mode=mode, encoding=encoding, delay=delay)


"""
FileName: api.py
Description: 
Time: 2020/9/11 16:42
Project: CiLog
Author: Shurui Gui
"""
import logging
from logging.handlers import SMTPHandler
from cilog.logger import CustomLogger, s2l
from cilog.handler import CustomFileHandler
from cilog.formatter import CustomFormatter, FileFormatter
import sys
import re
import time





def create_logger(name: str = 'CiLog',
                  file: str = None,
                  enable_mail: bool = False,
                  use_color: bool = True,
                  sub_print: bool = False,
                  use_printc: bool = False,
                  **kwargs) -> CustomLogger:
    """
    Create Custom Cilog for logging.

    Args:
        name : str : logger name
        file : str : File path
        enable_mail : bool : Default False
        use_color : bool : Signal for using colored info. Default True
        sub_print : bool : Substitute builtins.print [Advance Usage]
        **kwargs: Dict : Please refer to  the following options (hiden).

        file_mode : str - File open mode. Default: 'a'
        file_level : Literal['DEBUG', 'INFO', 'WARNING', 'IMPORTANT', 'ERROR', 'CRITICAL', 'MAIL'] - Default 'INFO'
        mail_level : Literal['DEBUG', 'INFO', 'WARNING', 'IMPORTANT', 'ERROR', 'CRITICAL', 'MAIL'] - Default 'MAIL'
        mail_setting : dir - Required if enable_mail == True
            {
                mailhost:   string or tuple - YourMailHost or (host, port),
                fromaddr:   string          - YourSenderAddress,
                toaddrs:    list(string)    - List of YourTargetAddresses,
                subject:    string          - Mail Subject,
                credentials:tuple           - (YourUsername, YourPassword),
                secure:     tuple           - () or (KeyfileName) or (KeyfileName, CertificatefileName)
                                                use the secure protocol (TLS),
                timeout:    float           - Default 1.0
            }
        stack_level : Literal['DEBUG', 'INFO', 'WARNING', 'IMPORTANT', 'ERROR', 'CRITICAL', 'MAIL'] - Default 'ERROR'
        msg_fmt : Dict{'DEBUG': debug_fmt, 'INFO': info_fmt, 'WARNING': warning_fmt, 'IMPORTANT': important_fmt,
        'ERROR': error_fmt, 'CRITICAL': critical_fmt, 'MAIL': mail_fmt} - Custom design massage format.
        Please refer to CustomFormatter and url: https://docs.python.org/3/library/logging.html#logrecord-attributes

    Return:
        logger : logging.Logger
    """

    kwargs['file_mode'] = kwargs.get('file_mode') or 'a'
    kwargs['file_level'] = s2l.get(kwargs.get('file_level') or 'INFO')
    kwargs['mail_level'] = s2l.get(kwargs.get('mail_level') or 'MAIL')
    kwargs['mail_setting'] = kwargs.get('mail_setting') or None
    kwargs['stack_level'] = s2l.get(kwargs.get('stack_level') or 'ERROR')
    kwargs['msg_fmt'] = kwargs.get('msg_fmt') or None


    # create logger
    logging.Logger.manager.setLoggerClass(CustomLogger) # set logging.Logger to CustomLogger
    logger: CustomLogger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # console handler and its formatter
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    cformatter = CustomFormatter(use_color=use_color,
                                 stack_level=kwargs['stack_level'],
                                 msg_fmt=kwargs['msg_fmt'])
    ch.setFormatter(cformatter)
    logger.addHandler(ch)

    # file handler
    if file:
        fh = CustomFileHandler(file, mode=kwargs['file_mode'])
        fh.setLevel(kwargs['file_level'])
        fformatter = FileFormatter(stack_level=kwargs['stack_level'], msg_fmt=kwargs['msg_fmt'])
        fh.setFormatter(fformatter)
        logger.addHandler(fh)

    if enable_mail:
        assert not(kwargs['mail_setting'] is None) and isinstance(kwargs['mail_setting'], dict), "Mail setting is needed."
        setting = kwargs['mail_setting']
        mh = SMTPHandler(setting.get('mailhost'), setting.get('fromaddr'), setting.get('toaddrs'),
                         setting.get('subject'), credentials=setting.get('credentials'),
                         secure=setting.get('secure'), timeout=setting.get('timeout'))
        mh.setLevel(kwargs['mail_level'])
        mformatter = FileFormatter(stack_level=kwargs['stack_level'], msg_fmt=kwargs['msg_fmt'])
        mh.setFormatter(mformatter)
        logger.addHandler(mh)

    if sub_print:
        logger.substitute_print()
        logging.StreamHandler.terminator = ''
        return


    return logger


def json_mail_setting(mail_setting):
    mail_setting["mailhost"] = tuple(mail_setting["mailhost"])
    mail_setting["credentials"] = tuple(mail_setting["credentials"])
    return mail_setting

class timeit(object):

    def __init__(self, time_name):
        self.time_name = time_name

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        print(f'{self.time_name}: {time.time() - self.start}s')






"""
FileName: formatter.py
Description: 
Time: 2020/9/11 16:41
Project: CiLog
Author: Shurui Gui
"""

import logging
from traceback import extract_stack, StackSummary



RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"
BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
COLORS = {
    'WARNING': YELLOW,
    'IMPORTANT': CYAN,
    'MAIL': CYAN,
    'INFO': WHITE,
    'DEBUG': BLUE,
    'CRITICAL': MAGENTA,
    'ERROR': RED
}


def bold_msg(message, use_color):
    if use_color:
        message = message.replace("$RESET", RESET_SEQ).replace("$BOLD", BOLD_SEQ)
    else:
        message = message.replace("$RESET", "").replace("$BOLD", "")
    return message


class CustomFormatter(logging.Formatter):

    def __init__(self, use_color, stack_level, msg_fmt=None, file=False):
        super().__init__("%(levelno)s: %(msg)s")
        self.use_color = use_color if not file else False
        self.stack_level = stack_level
        self._stack_prune = -8 if not file else -9
        self.msg_fmt = msg_fmt or \
            {
                'ORIGIN': "%(message)s",
                'DEBUG': "%(levelname)s: %(asctime)s : $BOLD%(message)s$RESET",
                'INFO': "%(levelname)s: $BOLD%(message)s$RESET",
                'WARNING': "%(levelname)s: %(filename)s - line %(lineno)d : $BOLD%(message)s$RESET",
                'IMPORTANT': "%(levelname)s: %(asctime)s : $BOLD%(message)s$RESET",
                'ERROR': "%(levelname)s: %(asctime)s - %(filename)s - line %(lineno)d : $BOLD%(message)s$RESET",
                'CRITICAL': "%(levelname)s: %(asctime)s - %(filename)s - line %(lineno)d : $BOLD%(message)s$RESET",
                'MAIL': "%(levelname)s: %(asctime)s : $BOLD%(message)s$RESET"
            }
        for key in self.msg_fmt.keys():
            self.msg_fmt[key] = bold_msg(self.msg_fmt[key], self.use_color)

    def format(self, record) -> str:

        # set for different levels
        format_orig = self._style._fmt
        self._style._fmt = self.msg_fmt[record.levelname]

        # preprocess \n
        line_count = 0
        for s in record.msg:
            if s == '\n':
                line_count += 1
                record.msg = record.msg[1:]
            else:
                break
        if record.msg:
            self._style._fmt = line_count * '\n' + self._style._fmt
        else:
            self._style._fmt = line_count * '\n'

        if record.levelno >= self.stack_level and record.levelno != logging.MAIL:
            record.stack_info = ''.join(StackSummary.from_list(extract_stack()[:self._stack_prune]).format())

        # make it colorful
        levelname = record.levelname
        if self.use_color and levelname in COLORS:
            levelname_color = COLOR_SEQ % (30 + COLORS[levelname]) + record.levelname + RESET_SEQ
            record.levelname = levelname_color

        self.datefmt = '%m/%d/%Y %I:%M:%S %p'
        result = logging.Formatter.format(self, record)

        self._style._fmt = format_orig
        record.levelname = levelname

        return result



class FileFormatter(CustomFormatter):

    def __init__(self, stack_level, msg_fmt=None):
        __file = True
        __use_color = False
        super().__init__(__use_color, stack_level, msg_fmt=msg_fmt, file=__file)


"""
FileName: logger.py
Description: 
Time: 2020/7/28 15:18
Project: CiLog
Author: Shurui Gui
"""

import logging
import re
import sys
import builtins
from functools import partial
from cilog.utils import CiLogStdOut
from pandas import DataFrame
import ast


logging.ORIGIN = 15
logging.IMPORTANT = 35
logging.TABLELIST = 36
logging.MAIL = 60

s2l = {
    'ORIGIN': logging.ORIGIN,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
    'IMPORTANT': logging.IMPORTANT,
    'TABLELIST': logging.TABLELIST,
    'MAIL': logging.MAIL,
    'CRITICAL': logging.CRITICAL,
    'ERROR': logging.ERROR,
    'NOTSET': logging.NOTSET
}

logging.addLevelName(logging.ORIGIN, 'ORIGIN')
logging.addLevelName(logging.IMPORTANT, 'IMPORTANT')
logging.addLevelName(logging.TABLELIST, 'TABLELIST')
logging.addLevelName(logging.MAIL, 'MAIL')


# ------------------------------ set new levels --------------------------------
class CustomLogger(logging.Logger):
    print = builtins.print

    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level=logging.NOTSET)

    def origin(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'ORIGIN'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.origin("Houston, we have a %s", "major disaster", exc_info=1)
        """
        if self.isEnabledFor(logging.ORIGIN):
            self._log(logging.ORIGIN, msg, args, **kwargs)

    def important(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'IMPORTANT'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.important("Houston, we have a %s", "major disaster", exc_info=1)
        """
        if self.isEnabledFor(logging.IMPORTANT):
            self._log(logging.IMPORTANT, msg, args, **kwargs)

    def table_fromlist(self, list_table: list, *args, **kwargs):
        """
        Log 'msg % args' with severity 'TABLELIST'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.table_fromlist(list_table, format)

        list_table: i.e.  [['model', 'dataset1', 'dataset2'],
                           ['m1', value11, value12],
                           ['m2', value21, value22]]
        """
        if self.isEnabledFor(logging.TABLELIST):
            format = 'markdown'
            if isinstance(list_table, str):
                match = re.match('!latex', list_table)
                if match:
                    format = 'latex'
                    list_table = list_table[match.span()[1]:]
                list_table = ast.literal_eval(list_table)

            table_dict = {key: [] for key in list_table[0]}
            for row_index in range(1, len(list_table)):
                row = list_table[row_index]
                for i, key in enumerate(table_dict.keys()):
                    table_dict[key].append(row[i])

            df = DataFrame(table_dict)

            if format == 'markdown':
                table = df.to_markdown(index=False)

            if format == 'latex':
                table = df.to_latex(index=False)

            self._log(logging.ORIGIN, table, args, **kwargs)
            self._log(logging.ORIGIN, '\n', args, **kwargs)


    def mail(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'MAIL'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.mail("Houston, we have a %s", "major disaster", exc_info=1)
        """
        if self.isEnabledFor(logging.MAIL):
            self._log(logging.MAIL, msg, args, **kwargs)

    def substitute_print(self):

        builtins.print = print_flush

        return CiLogStdOut(self)

    def restore_print(self):

        builtins.print = CustomLogger.print

        return


def print_flush(*values, sep=' ', end='\n', file=None, flush=True):
    sys.stdout.flush()
    CustomLogger.print(*values, sep=sep, end=end, file=file, flush=flush)

        # def log_print(*values, sep=' ', end='\n', file=sys.stdout, flush=False):
        #     if end != '\n' or file != sys.stdout or flush != False:
        #         builtins.print(values, sep=sep, end=end, file=file, flush=flush)
        #         return
        #
        #     for level in ['debug', 'important', 'warning', 'error', 'critical', 'mail']:
        #         match = re.match(str2mark(level), values[0])
        #         if match:
        #             getattr(self, level)(sep.join([str(value) for value in values])[match.span()[1]:])
        #             return
        #
        #     self.info(sep.join([str(value) for value in values]))
        #
        # return log_print








"""
FileName: launcher.py
Description: 
Time: 2020/9/12 13:15
Project: CiLog
Author: Shurui Gui
"""
import sys, subprocess, signal, os

sys_argv = sys.argv

def signal_process():
    # os.setsid()
    signal.signal(signal.SIGHUP, signal.SIG_IGN)

def launcher():
    subprocess.Popen(['python',
                      os.path.join(os.path.dirname(__file__), 'scripts.py'),
                      *sys_argv[1:]], preexec_fn=signal_process)




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cilog",
    version="1.1.1",
    author="Shurui Gui",
    author_email="citrinegui@gmail.com",
    entry_points={
        'console_scripts': [
            'cilog=cilog.launcher:launcher'
        ]
    },
    description="CiLog is a flexible integrated logging tool base on package logging that can substitute builtins.print"
                " directly.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CM-BF/CiLog/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)